/*************************************************************************/
/*									 */
/*	Main routine, c4.52						 */
/*	------------------						 */
/*      MODIFIED BY: D. Malerba                                          */
/*									 */
/*************************************************************************/


#include "defns1.i"
#include "types.i"


    /*  External data, described in extern.i  */

short		MaxAtt, MaxClass, MaxDiscrVal = 2;

ItemNo		MaxItem;

Description	*Item;

DiscrValue	*MaxAttVal;

char		*SpecialStatus;

String		*ClassName,
		*AttName,
		**AttValName,
		FileName = "DF",
                ParamFileName;

short		VERBOSITY = 0,
		TRIALS    = 10;

Boolean		GAINRATIO  = true,
		SUBSET     = false,
		BATCH      = true,
		UNSEENS    = false,
		PROBTHRESH = false;

ItemNo		MINOBJS   = 2,
		WINDOW    = 0,
		INCREMENT = 0;

float		CF = 0.25;

Tree		*Pruned;

Boolean		AllKnown = true;



    main(Argc, Argv)
/*  ----  */
    int Argc;
    char *Argv[];
{
    int o;
    extern char *optarg;
    extern int optind;
    Boolean  FirstTime=true;
    short    Best, BestTree();

    int      PruningMethod=StandardPrune;         /* Method of pruning used */

    PrintHeader("decision tree generator");

    /*  Process options  */

    while ( (o = getopt(Argc, Argv, "f:bupv:t:w:i:gsm:c:GRCM:102345IP")) != EOF )
    {
        if ( FirstTime )
	{
	    printf("\n    Options:\n");
	    FirstTime = false;
	}

	switch (o)
	{
	case 'f':   FileName = optarg;
		    printf("\tFile stem <%s>\n", FileName);
		    break;
	case 'b':   BATCH = true;
		    printf("\tWindowing disabled (now the default)\n");
		    break;
	case 'u':   UNSEENS = true;
		    printf("\tTrees evaluated on unseen cases\n");
		    break;
	case 'p':   PROBTHRESH = true;
		    printf("\tProbability thresholds used\n");
		    break;
	case 'v':   VERBOSITY = atoi(optarg);
		    printf("\tVerbosity level %d\n", VERBOSITY);
		    break;
	case 't':   TRIALS = atoi(optarg);
		    printf("\tWindowing enabled with %d trials\n", TRIALS);
                    Check(TRIALS, 1, 10000);
		    BATCH = false;
		    break;
	case 'w':   WINDOW = atoi(optarg);
	 	    printf("\tInitial window size of %d items\n", WINDOW);
                    Check(WINDOW, 1, 1000000);
		    BATCH = false;
		    break;
	case 'i':   INCREMENT = atoi(optarg);
		    printf("\tMaximum window increment of %d items\n",
			   INCREMENT);
                    Check(INCREMENT, 1, 1000000);
		    BATCH = false;
		    break;
	case 'g':   GAINRATIO = false;
		    printf("\tGain criterion used\n");
		    break;
	case 's':   SUBSET = true;
		    printf("\tTests on discrete attribute groups\n");
		    break;
	case 'm':   MINOBJS = atoi(optarg);
		    printf("\tSensible test requires 2 branches with >=%d cases\n",
			    MINOBJS);
                    Check(MINOBJS, 1, 1000000);
		    break;
	case 'c':   CF = atof(optarg);
		    printf("\tPruning confidence level %g%%\n", CF);
                    Check(CF, 0, 100);
		    CF /= 100;
		    break;
        case 'G':   printf("\tReduced Error Grafting\n");
                    PruningMethod=GRA;
                    break;
        case 'R':   printf("\tReduced error pruning \n");
                    PruningMethod=REP;  
                    break;
        case 'C':   printf("\tCritical value pruning\n");
                    PruningMethod=CVP;
                    break;
        case '3':   printf("\tError complexity pruning with cross-validation and1SE rule \n");
                    PruningMethod=SE3;
                    break;
        case '2':   printf("\tError complexity pruning with cross-validation\n");
                    PruningMethod=SE2;
                    break;
        case '5':   printf("\tError complexity pruning (NEW) with cross-validation and1SE rule \n");
                    PruningMethod=SE5;
                    break;
        case '4':   printf("\tError complexity pruning (NEW) with cross-validation\n");
                    PruningMethod=SE4;
                    break;
        case '1':   printf("\tError complexity pruning with test set and 1SE rule\n");
                    PruningMethod=SE1;
                    break;
        case '0':   printf("\tError complexity pruning with test set\n");
                    PruningMethod=SE0;
                    break;
        case 'M':   ParamFileName = optarg;
                    printf("\tMinimum error pruning \n");
                    printf("\tParameters file name <%s>\n",ParamFileName);
                    PruningMethod=MEP;
                    break;
        case 'I':   printf("\tIterative growing and pruning algorithm\n");
                    PruningMethod=IGP;
                    break;
        case 'P':   printf("\tPessimistic error pruning method \n");
                    PruningMethod=PEP;
                    break;
	case '?':   printf("unknown option\n");
		    exit(1);
	}
    }

    /*  Initialise  */

    GetNames();
    switch (PruningMethod)
      {
        case GRA              :
        case REP              :
        case SE0              :
        case SE1              : 
        case CVP              :
        case MEP              : GetData(".grow");
                                printf("\nRead %d cases (%d attributes) from %s.grow\n",MaxItem+1,MaxAtt+1,FileName);
                                break;
        case PEP              :
        case SE2              :
        case SE3              :
        case SE4              :
        case SE5              : GetData(".train"); 
                                printf("\nRead %d cases (%d attributes) from %s.train\n",MaxItem+1,MaxAtt+1,FileName);
                                break;
        case IGP              : GetData(".train1"); 
                                printf("\nRead %d cases (%d attributes) from %s.train1\n",MaxItem+1,MaxAtt+1,FileName);
                                break;
        case StandardPrune    : GetData(".data"); 
                                printf("\nRead %d cases (%d attributes) from %s.data\n",MaxItem+1,MaxAtt+1,FileName);
                                break;
      }

    /*  Build decision trees  */

    if ( BATCH )
    {
	TRIALS = 1;
	OneTree(PruningMethod);
	Best = 0;
    }
    else
    {
	Best = BestTree();
    }

    /*  Soften thresholds in best tree  */

    if ( PROBTHRESH )
    {
	printf("Softening thresholds");
	if ( ! BATCH ) printf(" for best tree from trial %d", Best);
	printf("\n");
	SoftenThresh(Pruned[Best]);
	printf("\n");
	PrintTree(Pruned[Best]);
    }

    /*  Save best tree  */

    if ( BATCH || TRIALS == 1 )
    {
	printf("\nTree saved\n");
    }
    else
    {
	printf("\nBest tree from trial %d saved\n", Best);
    }
    SaveTree(Pruned[Best], ".tree");

    /*  Evaluation  */

    printf("\n\nEvaluation on training data (%d items):\n", MaxItem+1);
    if (PruningMethod==IGP) EvaluateForIGP(false, Best);
    else Evaluate(false, Best);

    if ( UNSEENS )
    {   
        GetData(".test");
        printf("\nEvaluation on test data (%d items):\n", MaxItem+1);
        if (PruningMethod==IGP) EvaluateForIGP(false, Best);
        else Evaluate(false, Best);
        Verbosity(1)
          {
            if (PruningMethod==IGP) EvaluateAndSaveForIGP(Best);
            else EvaluateAndSave(Best);
            printf("\n\nSAVED STATISTICS ON tree001, prune001 AND MatDist001 FILES\n\n");
          }
    }   
}
