/*************************************************************************/
/*                                                              	 */
/*                       Cross validation routines       	         */
/*  --------------------------------------------------------------	 */
/*                       MODIFIED BY D. Malerba                           */
/*                                                              	 */
/*************************************************************************/


#include "defns1.i"
#include "types.i"
#include "extern.i"
#include <string.h>
#include <math.h>

Tree      *GrownTrees;



/**************************************************************************/
/*                                                                        */
/*       DESCRPTION:
/*             For each tree in the Family estimate the error rate        */
/*             and the size.                                              */
/*               INPUT   ParamValues -> values of the parameters          */
/*                                      corresponding to the trees        */
/*                       NoTrees     -> Number of trees to evaluate       */ 
/*               OUTPUT: Error       -> error estimated for each tree     */
/*                                                                        */
/*       AUTHOR: D. Malerba                                               */
/**************************************************************************/

CrossValidationNew(Error,ParamValues,NoTrees)
  float  *Error,*ParamValues;
  int    NoTrees;

  {
    int     i,j,v;
    Tree    *PrunedTrees,*ToPrune,*UpdatedTrees,CopyTree(),*TreeToPrune;
    int     NodeNumber,MaxNodeNumber=1,DimTrainSet,count;
    float   TempErrorRate,TotalErrorRate=0.0;
    float   MinAlfaValue=1000.0,TempErrorRate2;
    float   *AlfaValori[CROSSNUMBER];
    Tree    *PrunedAlfa[CROSSNUMBER];
    Boolean FindAlfaTree();


    GrownTrees  =(Tree *) calloc(CROSSNUMBER,sizeof(Tree));
    PrunedTrees =(Tree *) calloc(1,sizeof(Tree));
    ToPrune     =(Tree *) calloc(1,sizeof(Tree));
    UpdatedTrees=(Tree *) calloc(CROSSNUMBER,sizeof(Tree));
     
   
    if (NoTrees>= 0) 
      {
        GrownValidationTrees(UpdatedTrees);
        ForEach(i,1,CROSSNUMBER)
          {   
             printf("\n\n\nAlfa Values for CROSS VALIDATION TREE NUMBER %d\n\n\n",i);
             AlfaValori[i-1] = (float *) calloc(1000,sizeof(float));
             PrunedAlfa[i-1] = (Tree *) calloc(1000,sizeof(Tree));
             ToPrune[0]= CopyTree(GrownTrees[i-1]);
             PrunedAlfa[i-1][0]=CopyTree(UpdatedTrees[i-1]);
             PrunedAlfa[i-1][1]= CopyTree(UpdatedTrees[i-1]);
             AlfaValori[i-1][0]=0.0;
             count=2;
             MaxItem=GrownTrees[i-1]->Items;
             NodeNumber=1;
             while (LeafCount(PrunedAlfa[i-1][count-1])!=1)
               {
                 UpdateAlfaTree(PrunedAlfa[i-1][count-1],ToPrune[0],&MinAlfaValue);
                 if (MinAlfaValue >0)
                 {
                  AlfaValori[i-1][count-1]=MinAlfaValue;  
                  PruneAlfa(MinAlfaValue,PrunedAlfa[i-1][count-1],ToPrune[0],NodeNumber,&MaxNodeNumber);
                  printf("TREE NUMBER %d \t Alfa=%.6f\n",count-1,MinAlfaValue);   
                  PrintTree(PrunedAlfa[i-1][count-1]);
                  PrunedAlfa[i-1][count]=CopyTree(PrunedAlfa[i-1][count-1]);
                  count++;
                  if (count >1000)
                    {
                      printf("\n\nError ->Family of pruned trees too large\n\n");
                      exit(1);
                    }
                 }
                 else
                   PruneAlfa(MinAlfaValue,PrunedAlfa[i-1][count-1],ToPrune[0],NodeNumber,&MaxNodeNumber);
                 MaxNodeNumber=1;
                 MinAlfaValue=1000.0;
               }
              ReleaseTree(ToPrune[0]);
             ReleaseTree(GrownTrees[i-1]);
          }  
    
        ForEach(i,0,NoTrees)
          {
            count=1; 
            ForEach(v,1,CROSSNUMBER)
              {                
                while ((AlfaValori[v-1][count]< ParamValues[i])&&(AlfaValori[v-1][count]>AlfaValori[v-1][count-1])) count++;
                if (AlfaValori[v-1][count-1]< AlfaValori[v-1][count])
                  {                                      
                    EstimateValidationTrees(PrunedAlfa[v-1][count-1],&TempErrorRate,v);
                    TempErrorRate*=(1.-((ParamValues[i]-AlfaValori[v-1][count-1])/(AlfaValori[v-1][count]-AlfaValori[v-1][count-1]))); 
                    EstimateValidationTrees(PrunedAlfa[v-1][count],&TempErrorRate2,v); 
                    TempErrorRate2*=(ParamValues[i]-AlfaValori[v-1][count-1])/(AlfaValori[v-1][count]-AlfaValori[v-1][count-1]);
                    TotalErrorRate+=TempErrorRate+TempErrorRate2;
                    count=1;
                  } 
                else
                  {  
                    EstimateValidationTrees(PrunedAlfa[v-1][count-1],&TempErrorRate,v);
                    TotalErrorRate+=TempErrorRate;
                    count=1;
                  }
               }
            Error[i]=TotalErrorRate/CROSSNUMBER;
            TotalErrorRate=0.0;
          }
        /* ForEach(v,1,CROSSNUMBER) 
          {
            EstimateValidationTrees(PrunedAlfa[v-1][0],&TempErrorRate,v);
            TotalErrorRate+=TempErrorRate;
          }
          Error[0]=TotalErrorRate/CROSSNUMBER;*/
 
         ForEach(i,0,CROSSNUMBER-1)
            {
               v=0;
            while (PrunedAlfa[i][v] != Nil )
              {                
                
                ReleaseTree(PrunedAlfa[i][v]); 
                ++v;
               }
             } 
         ForEach(i,0,CROSSNUMBER-1)
            cfree(AlfaValori[i]);
   }        
}





CrossValidation(Error,ParamValues,NoTrees)
  float  *Error,*ParamValues;
  int    NoTrees;

  {
    int     i,j,v;
    Tree    *PrunedTrees,*ToPrune,*UpdatedTrees,CopyTree();
    int     NodeNumber,MaxNodeNumber=1,DimTrainSet;
    float   TempErrorRate,TotalErrorRate=0.0;
    Boolean FindAlfaTree();

    GrownTrees  =(Tree *) calloc(CROSSNUMBER,sizeof(Tree));
    PrunedTrees =(Tree *) calloc(1,sizeof(Tree));
    ToPrune     =(Tree *) calloc(1,sizeof(Tree));
    UpdatedTrees=(Tree *) calloc(CROSSNUMBER,sizeof(Tree));

    if (NoTrees-1 >= 0) 
      {
        GrownValidationTrees(UpdatedTrees);
        ForEach(i,0,NoTrees-1)
          {
            ForEach(v,1,CROSSNUMBER)
              {
                ToPrune[0]=CopyTree(GrownTrees[v-1]);
                PrunedTrees[0]=CopyTree(UpdatedTrees[v-1]);
                DimTrainSet=(int) GrownTrees[v-1]->Items;
                Verbosity(1)
                  printf("\n\n\nNUMBER OF THE PRUNED VALIDATION TREE %d.%d \tAlfa: %.6f \n\n",i+1,v,(float)sqrt(ParamValues[i]*ParamValues[i+1]));
                FindAlfaTree(PrunedTrees[0],ToPrune[0],(float)sqrt(ParamValues[i]*ParamValues[i+1]),DimTrainSet,NodeNumber,&MaxNodeNumber);
                Verbosity(1)
                  PrintTree(PrunedTrees[0]);
                EstimateValidationTrees(PrunedTrees[0],&TempErrorRate,v);
                Verbosity(1)
                  {
                    printf("\n\n\n \t\t\tERROR ESTIMATE FOR TREE NUMBER %d.%d:",i+1,v); 
                    printf("\n \t\t\tErrorRate= %.2f\n",TempErrorRate);
                  }
                TotalErrorRate+=TempErrorRate;
                MaxNodeNumber=1;
                ReleaseTree(PrunedTrees[0]); 
                ReleaseTree(ToPrune[0]); 
              } 
            Error[i]=TotalErrorRate/CROSSNUMBER;
            TotalErrorRate=0.0; 
          }
        ForEach(i,1,CROSSNUMBER)
          {
            ReleaseTree(GrownTrees[i-1]); 
            ReleaseTree(UpdatedTrees[i-1]);
          }
      } 
  }







/****************************************************************************/
/*                                                                          */
/*     DESCRIPTION:                                                         */
/*      Find the tree T that minimize the cost-complexity measure for       */
/*      the value of alfa equal to Alfa                                     */
/*     AUTHOR: D. Malerba                                                   */
/*                                                                          */
/****************************************************************************/

Boolean FindAlfaTree(T,R,Alfa,DimTrainSet,NodeNumber,MaxNodeNumber)
  Tree   T,R;
  float  Alfa;
  int    DimTrainSet,NodeNumber,*MaxNodeNumber;
  
  {
    Boolean Changed;
    int     PreviousNodeNumber,v,i,LeafCount();
    float   SubTreeErrors=0.0,Decrease;

    Changed=false;
    if (T->NodeType != 0)
      {
        PreviousNodeNumber=NodeNumber;
        ForEach(i,1,T->Forks)
          {
            (*MaxNodeNumber)++;
            NodeNumber= (*MaxNodeNumber);
            FindAlfaTree(T->Branch[i],R->Branch[i],Alfa,DimTrainSet,NodeNumber,MaxNodeNumber);
            SubTreeErrors+=R->Branch[i]->Errors;
          }
        NodeNumber=PreviousNodeNumber;
        
        Decrease=(R->Errors - T->Errors)/DimTrainSet + Alfa*(LeafCount(R)-1);
 
        if (Decrease>=0)
          {
            Verbosity(1)
              printf("PRUNED NODE NUMBER: %d \tR(t): %.2f \tR(Tt): %.2f \tAlfa: %.7f\n",NodeNumber,(T->Errors)/DimTrainSet,(R->Errors)/DimTrainSet,Alfa);
            ForEach(v,1,T->Forks)
              {
                ReleaseTree(T->Branch[v]);
                T->Branch[v] = Nil;
                ReleaseTree(R->Branch[v]);
                R->Branch[v]=Nil;
              };
            R->Errors=T->Errors;
            T->NodeType=0;
            R->NodeType=0;
            Changed = true;
          }
        else R->Errors=SubTreeErrors;
      }
    else (*MaxNodeNumber)--;
    return Changed;
  } 

/****************************************************************************/
/*                                                                          */
/*       DESCRIPTION: Grown CROSSNUMBER trees for validation                */
/*       AUTHOR: D. Malerba                                                 */
/*                                                                          */
/****************************************************************************/

GrownValidationTrees(UpdatedTrees)
  Tree *UpdatedTrees;

  {
    int  i;
    char ext[6],num[6],extTemp[6],*strcpy(),*strcat();
    Tree CopyTree(),FormTree();

    strcpy(ext,".xt\0");
    strcpy(extTemp,ext);
    Verbosity(1)
      printf("\n\n\n\n \t\t\tCROSS VALIDATION TREES:\n\n"); 
    ForEach(i,1,CROSSNUMBER)
      {
        itoa(i,num);
        strcat(extTemp,num);
        FreeItem();
        GetData(extTemp);
        GrownTrees[i-1]=FormTree(0,MaxItem);
        UpdatedTrees[i-1]=CopyTree(GrownTrees[i-1]);
        UpdateTree(UpdatedTrees[i-1],0,MaxItem);
        Verbosity(1)
          {
             printf("\n\n\nVALIDATION TREE NUMBER %d\n\n",i);
             PrintTree(GrownTrees[i-1]);
          }
        strcpy(extTemp,ext); 
      } 
  }






/******************************************************************************/
/*                                                                            */
/*      DESCRPTION: Estimate error rate on test sets of validation number i   */
/*                                                                            */
/******************************************************************************/

EstimateValidationTrees(T,ErrorRate,i)
  Tree  T;
  float *ErrorRate;
  int   i;

  {
    char   ext[6],num[6],extTemp[6],*strcpy(),*strcat();
    ItemNo Misclassify();

    strcpy(ext,".xs\0");
    strcpy(extTemp,ext);
    itoa(i,num);
    strcat(extTemp,num);
    FreeItem();
    GetData(extTemp);
    *ErrorRate=Misclassify(T)/(MaxItem+1.0);
  }



/*****************************************************************************/
/*                                                                           */
/*       DESCRPTION: Convert an integer n to  string s                       */
/*       AUTHOR: D. Malerba                                                  */
/*                                                                           */
/*****************************************************************************/

itoa(n,s)
  int  n;
  char s[3];

  {
    int i,sign;

    if ((sign=n)<0) n= -n;                    /* record sign */
    i=0;
    do
      {                                       /* generate digits in reverse order */
        s[i++]=n % 10 + '0';                  /* get next digit                   */
      }
    while ((n/=10)>0);                        /* delete it                        */
    if (sign<0) s[i++]='-';
    s[i]='\0';
    reverse(s);
  }



/*******************************************************************************/
/*                                                                             */
/*       DESCRIPTION: reverse string s in place                                */
/*       AUTHOR: D. Malerba                                                    */
/*                                                                             */
/*******************************************************************************/

reverse(s)
  char s[3];

  {
    int c,i,j;

    for (i=0, j=strlen(s)-1; i<j; i++, j--)
      {
        c=s[i];
        s[i]=s[j];
        s[j]=c;
      }
  }



/****************************************************************************/
/*                                                                          */
/*       DESCRIPTION: free mem allocated for Item[]                         */
/*       AUTHOR: D. Malerba                                                 */
/*                                                                          */
/****************************************************************************/

FreeItem()

  {
    ItemNo i;

    ForEach(i,0,MaxItem)
      cfree(Item[i]);
    free(Item);
  }




          
